# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), The AiiDA-CP2K authors.                                      #
# SPDX-License-Identifier: MIT                                                #
# AiiDA-CP2K is hosted on GitHub at https://github.com/aiidateam/aiida-cp2k   #
# For further information on the license, see the LICENSE.txt file.           #
###############################################################################
"""AiiDA-CP2K input plugin"""

from __future__ import absolute_import
from copy import deepcopy

import six


class Cp2kInput:  # pylint: disable=old-style-class
    """Transforms dictionary into CP2K input"""

    DISCLAIMER = "!!! Generated by AiiDA !!!"

    def __init__(self, params=None):
        """Initializing Cp2kInput object"""
        if not params:
            self._params = {}
        else:
            # always make a full copy to avoid that add_keyword() changes the
            # passed-in dictionary
            self._params = deepcopy(params)

    def add_keyword(self, kwpath, value):
        """
        Add a value for the given keyword.

        Args:
            kwpath: Can be a single keyword, a path with `/` as divider for sections & key,
                    or a sequence with sections and key
            value: the value to set the given key to
        """

        if isinstance(kwpath, six.string_types):
            kwpath = kwpath.split("/")

        Cp2kInput._add_keyword(kwpath, value, self._params)

    def render(self):
        output = [self.DISCLAIMER]
        self._render_section(output, deepcopy(self._params))
        return u"\n".join(output)

    @staticmethod
    def _add_keyword(kwpath, value, params):
        """Add keyword in given nested dictionary"""

        if len(kwpath) == 1:  # key/value for the current level
            params[kwpath[0]] = value
            return

        if kwpath[0] not in params.keys():  # create an empty section if necessary
            params[kwpath[0]] = {}

        Cp2kInput._add_keyword(kwpath[1:], value, params[kwpath[0]])

    @staticmethod
    def _render_section(output, params, indent=0):
        """
        It takes a dictionary and recurses through.

        For key-value pair it checks whether the value is a dictionary
        and prepends the key with & (CP2K section)
        It passes the valued to the same function, increasing the indentation
        If the value is a list, I assume that this is something the user
        wants to store repetitively
        eg:
            dict['KEY'] = ['val1', 'val2']
            ===>
            KEY val1
            KEY val2

            or

            dict['KIND'] = [{'_': 'Ba', 'ELEMENT':'Ba'},
                            {'_': 'Ti', 'ELEMENT':'Ti'},
                            {'_': 'O', 'ELEMENT':'O'}]
            ====>
                  &KIND Ba
                     ELEMENT  Ba
                  &END KIND
                  &KIND Ti
                     ELEMENT  Ti
                  &END KIND
                  &KIND O
                     ELEMENT  O
                  &END KIND
        """

        if six.PY2:
            from collections import Mapping, Sequence
        else:
            from collections.abc import Mapping, Sequence  # pylint: disable=import-error, no-name-in-module
        for key, val in sorted(params.items()):
            # keys are not case-insensitive, ensure that they follow the current scheme
            if key.upper() != key:
                raise ValueError("keyword '{}' not upper case".format(key))

            if key.startswith(("@", "$")):
                raise ValueError("CP2K preprocessor directives not supported")

            if isinstance(val, Mapping):
                line = "{}&{}".format(' ' * indent, key)
                if "_" in val:  # if there is a section parameter, add it
                    line += " {}".format(val.pop('_'))
                output.append(line)
                Cp2kInput._render_section(output, val, indent + 3)
                output.append('{}&END {}'.format(' ' * indent, key))

            elif isinstance(val, Sequence) and not isinstance(val, six.string_types):
                for listitem in val:
                    Cp2kInput._render_section(output, {key: listitem}, indent)

            elif isinstance(val, bool):
                val_str = '.TRUE.' if val else '.FALSE.'
                output.append('{}{} {}'.format(' ' * indent, key, val_str))

            else:
                output.append('{}{} {}'.format(' ' * indent, key, val))
